
let DynamoParam = require('./dynamo_param.js');

function DynamoObject(data, source) {
    this.params = {};
    this.source = source;
    /**
     * {
        tablename : tablename,
        region : region,
        count: INTEGER,
        sorts: [array of string == keys of the model which are the sort keys to use]
    }
     */
    this.excludes = ['params', 'source', 'modified', 'dynamo'];
    this.source.contains_sorts = false;
    this.modified_index = {};
    this.modified = [];
    this.delete_index = [];
    if (this.source.sorts) {
        let it = this;
        for (let i = 0; i < this.source.sorts.length; i++) {
            let key = it.source.sorts[i];
            this.source.contains_sorts = !(!data[key]);
        }
    }

    if (!this.source.contains_sorts) return console.log("Error - DynamoObject Missing Some or All Sort Keys", this.source.sorts);

    if (typeof data === 'string') ParseDataString(data, this);
    if (typeof data === 'object') ParseDataObject(data, this);

    this.reApply = function (data) {
        if (typeof data === 'string') ParseDataString(data, this);
        if (typeof data === 'object') ParseDataObject(data, this);
    };

    this.set = function (tag, value, isObject) {
        if (isObject && isObject === true) {

        } else {
            if (this.source.sorts.indexOf(tag) > -1) return;
            let param = this.params[tag];
            this[tag] = value;
            let t = this;
            if (!value) return console.log('No Value Given FOr Dynamo Object Set', this.instance_id);
            if (!param) {
                console.log('Setting Value For', tag, ' which does not exist, but is a Parameter');
                this.params[tag] = new DynamoParam(value);
            } else param.set(value);
            UpMod(tag);
            function UpMod() {
                let index = t.modified_index[tag];
                if (!index) {
                    t.modified_index[tag] = {n_modified: 1};
                    t.modified.push(tag);
                }
                else index.n_modified += 1;
            }
        }
    };

    this.delete = function (value, callback) {
        let tag, perform, dynamo = this.dynamo, it = this;
        if (value.constructor === Array) HandleArrayValues(value, callback);
        else switch (typeof value) {
            case 'string':
                tag = value;
                HandleStringValue(tag, callback);
                break;
            case 'object':
                tag = value.tag;
                perform = value.perform;
                if (tag.constructor === Array) HandleArrayValues(tag, callback);
                if (typeof tag === 'string') HandleStringValue(tag, callback);
                break;
            default: return (callback) ? callback('Resulting '+value+"not supported") : null;
        }
        if (perform && dynamo) return new Promise(function (result) {
            dynamo.updateItem(it.updateParams('DELETE'), function (err, done) {
                if (!err) it.clearModifies();
                delete it.params[tag];
                delete it[tag];
                return result((err) ? Promise.reject(err) : Promise.resolve());
            })
        });
        function HandleStringValue(tagged, callback) {
            if (it.source.sorts.indexOf(tagged) > -1) return (callback) ? callback('Cannot Remove A Sorting Key') : null;
            let mod = it.modified.indexOf(tagged), mod_ind = it.modified_index[tagged], del = it.delete_index.indexOf(tagged);
            if (del === -1) it.delete_index.push(tagged);
            if (mod > -1) delete it.modified[mod];
            if (mod_ind) delete it.modified_index[tagged];
        }

        function HandleArrayValues(tags, callback) {
            for (let i = 0; i < tags.length; i++) {
                HandleStringValue(tags[i]);
            }
        }
    };

    this.searchParams = function () {
        let Keys = {};
        let sorts = this.source.sorts;
        let it = this;
        for (let i = 0; i < sorts.length; i++) {
            let key = sorts[i];
            let param = it.params[key];
            Keys[key] = param.dynamoObject();
        }
        return {
            TableName : it.source.tablename,
            Keys : Keys
        }
    };

    this.update = function(type) {
        let it = this;
        if (!type) return new Promise(function (result) {
            let update = it.updateParams(type);
            it.dynamo.updateItem(update, function (err, upd) {
                if(!err) it.clearModifies(true);
                return result((err) ? Promise.reject(err) : Promise.resolve());
            });
        });
        else if (type.indexOf('&') > -1) {
            let split = type.split('&');
            let final;
            split.forEach(function (value) {
                let pack = it.updateParams(value);
                if (!final) final = pack;
                else {
                    let vals = pack.ExpressionAttributeValues,
                        names = pack.ExpressionAttributeNames,
                        newNames = Object.assign(final.ExpressionAttributeNames, names),
                        newExp = final.UpdateExpression + " " + pack.UpdateExpression;
                    final.ExpressionAttributeValues = Object.assign(final.ExpressionAttributeValues, vals);
                    final.ExpressionAttributeNames = newNames;
                    final.UpdateExpression = newExp;
                }
            });
            return new Promise(function (result) {
                it.dynamo.updateItem(final, function (err, upd) {
                    if(!err) it.clearModifies(true);
                    return result((err) ? Promise.reject(err) : Promise.resolve());
                })
            })
        }
        else return new Promise(function (result) {
            let update = it.updateParams(type);
            it.dynamo.updateItem(update, function (err, upd) {
                if(!err) {
                    switch (type) {
                        case 'SET' :
                            it.clearModifies(); break;
                        case 'DELETE' :
                            it.clearDeletes(); break;
                        default: break;
                    }
                }
                return result((err) ? Promise.reject(err) : Promise.resolve());
            })
        })
    };

    this.batchUpdate = function () {
        let it = this,
            dels = it.delete_index,
            mods = it.modified,
            delsLength = dels.length,
            modsLength = mods.length;
        return new Promise(function (result) {
            if (delsLength === 0 && modsLength > 0) it.update('DELETE').then(result).catch(result);
            else if (delsLength > 0 && modsLength === 0) it.update('SET').then(result).catch(result);
            else it.update('SET&DELETE').then(result).catch(result)
        });

    };

    this.deleteParams = function () {
        return this.updateParams('DELETE');
    };

    this.updateParams = function (action) {
        let allKeys;
        let doAction = "SET";
        if (!action) {
            allKeys = this.modified;
        } else {
            switch (action) {
                case 'SET':
                    allKeys = this.modified;
                    break;
                case 'DELETE' || 'REMOVE':
                    allKeys = this.delete_index;
                    doAction = "REMOVE";
                    break;
                default:return;
            }
        }
        let update = {},
            UpdateExpression = "",
            ExpressionAttributeValues = {},
            ExpressionAttributeNames = {},
            Keys = {},
            allKeysLength = allKeys.length,
            it = this,
            sorts = this.source.sorts,
            sortsLength = sorts.length;
        let sortCount = 0;
        let attCount = 0;
        for (let i = 0; i < allKeysLength; i++) {
            let tag = allKeys[i];
            if (!tag || typeof tag !== 'string') return console.log('No Such Tag');
            let param = it.params[tag];
            if (param) {
                let exptag, valtag;
                if (doAction === 'SET') {
                    if (tag.indexOf('_') > -1) {
                        let tagers = tag.split("_"), tag1 = tagers[0].substring(0,2), tag2 = tagers[1].substring(0,2), final =(tag1+tag2);
                        exptag = ("#"+final.toUpperCase());
                        valtag = ":"+final;
                    } else {
                        exptag = "#"+tag.substring(0,2).toUpperCase();
                        valtag = ":"+tag.substring(0,2);
                    }
                    ExpressionAttributeValues[valtag] = param.dynamoObject();
                    ExpressionAttributeNames[exptag] = tag;
                    if (i < allKeys.length - 1) {
                        UpdateExpression += " "+exptag+" = "+valtag+",";
                    } else {
                        UpdateExpression += " "+exptag+" = "+valtag;
                    }
                } else if (doAction === "REMOVE") {
                    if (tag.indexOf('_') > -1) {
                        let tagers = tag.split("_"), tag1 = tagers[0].substring(0,2), tag2 = tagers[1].substring(0,2), final =(tag1+tag2);
                        exptag = ("#"+final.toUpperCase());
                        valtag = ":"+final;
                    } else {
                        exptag = "#"+tag.substring(0,2).toUpperCase();
                        valtag = ":"+tag.substring(0,2);
                    }
                    ExpressionAttributeValues[valtag] = param.dynamoObject();
                    if (i < allKeys.length - 1) {
                        UpdateExpression += " "+tag+",";
                    } else {
                        UpdateExpression += " "+tag;
                    }
                }
                attCount += 1;
            }
        }
        for (let i = 0; i < sorts.length; i++) {
            let tag = sorts[i];
            let param = it.params[tag];
            if (param) {
                Keys[tag] = param.dynamoObject();
                sortCount += 1;
            }
        }
        if (sortCount !== sorts.length) return console.log('Invalid Sorter', sortCount, sortsLength);
        if (attCount !== allKeys.length) return console.log('Invalid Attributes', attCount, allKeysLength);
        update.Key = Keys;
        switch(doAction) {
            case 'SET' :
                update.ExpressionAttributeValues = ExpressionAttributeValues;
                update.ExpressionAttributeNames = ExpressionAttributeNames;
                break;
            default:
                break;
        }
        update.UpdateExpression = doAction+UpdateExpression;
        update.TableName = this.source.tablename;
        return update;
    };

    this.clearModifies = function (all) {
        this.modified = [];
        this.modified_index = {};
        if (all) this.clearDeletes();
    };
    this.clearDeletes = function () {
        let it = this;
        this.delete_index.forEach(function (value) {
            delete it.params[value];
            delete it[value];
        });
        this.delete_index = [];
    };

    function ParseDataString(string, it) {
        let object = JSON.parse(string);
        ParseDataObject(object, it);
    }
    function ParseDataObject(object, it) {
        let keys = Object.keys(object);
        if (!source) source = {count: keys.length};
        else source.count = keys.length;
        for (let i = 0; i < keys.length; i++) {
            let current = keys[i];
            let val = object[current];
            let param = new DynamoParam(val);
            it[current] = param.value;
            it.params[current] = param;
        }
    }

}


module.exports = DynamoObject;

